@@ -3,6 +3,7 @@ from collections import defaultdict
 from concurrent.futures import Future, ProcessPoolExecutor, as_completed
 from pathlib import Path
 from typing import Dict, Iterator, List, Set, Tuple, Union
+import itertools
 
 import networkx as nx
 import pandas as pd
@@ -161,12 +162,21 @@ def get_all_motif_MSTs_generator(motif_MST: nx.Graph, max_n_mutated_residues: in
                 
                 yield mutated_graph
     
-def get_all_pairs_of_residues_in_motif_MST(motif_MST: nx.Graph) -> Dict[Tuple[str,str], Residue_pair_data]:
+def get_all_pairs_of_residues_in_motif_MST(motif_MST: nx.Graph, order: str = 'dfs') -> Dict[Tuple[str,str], Residue_pair_data]:
     """
+    Returns all the residue pairs in motif_MST in an ordered manner such that, for all i, residue pairs between 0 and i form a connected graph. The order can 
+    be that of a Depth First Search (dfs) of Breadth First Search (bfs). If the residue pairs were randomly ordered, we wouldn't be able to do the connectivity check 
+    in update_map_of_PDBs_with_all_residue_pairs, because the logic requires that new residue pairs are connected to the previous nodes.
     """
+    MST_edges: List[Tuple[str, str]] # Ex: [('A1G', 'A3K'), ('A3K', 'A8N'), ...]
+    if order == 'dfs':
+        MST_edges = list(nx.dfs_edges(motif_MST))
+    elif order == 'bfs':
+        MST_edges = list(nx.bfs_edges(motif_MST))
+    else:
+        raise ValueError(f" The 'order' parameter must be either 'dfs' or 'bfs', but {order} was given.")
+    
     all_pairs_of_residues_in_motif_MST: Dict[Tuple[str,str], Residue_pair_data] = {}
-
-    MST_edges: List[Tuple[str, str]] = list(motif_MST.edges) # Ex: [('A1G', 'A3K'), ('A3K', 'A8N'), ...]
     for pair_of_full_residue_IDs in MST_edges:
         pair_of_full_residue_IDs = get_sorted_2_tuple(pair_of_full_residue_IDs) # Ordering is needed to make sure we only calculate each pair once, i.e avoid calculating the results of (A,B) and (B,A), as they are identical.
         
@@ -285,8 +295,11 @@ def get_PDBs_that_contain_the_residue_pair(
         
     return PDBs_that_contain_the_residue_pair
 
+def residue_pair_is_connected_to_visited_nodes(residue_pair: Tuple[str, str], visited_nodes: Set[str]) -> bool:
+    return residue_pair[0] in visited_nodes or residue_pair[1] in visited_nodes
+
 def update_map_of_PDBs_with_all_residue_pairs(
-        map_of_PDBs_with_all_residue_pairs: Dict[str, List[Tuple[str,str]]], new_residue_pair_data: Dict[str, List[Tuple[str,str]]]
+        map_of_PDBs_with_all_residue_pairs: Dict[str, List[Tuple[str,str]]], new_residue_pair_data: Dict[str, List[Tuple[str,str]]], visited_nodes_map: Dict[str, Set[str]]
     ) -> None:
     """
     ...
@@ -294,10 +307,21 @@ def update_map_of_PDBs_with_all_residue_pairs(
     updated_set_of_PDBs_with_all_residue_pairs = map_of_PDBs_with_all_residue_pairs.keys() & new_residue_pair_data.keys() # & = set intersection
     
     for PDB_ID in list(map_of_PDBs_with_all_residue_pairs.keys()):
+        # Remove PDBs that no longer have all the residue pairs
         if PDB_ID not in updated_set_of_PDBs_with_all_residue_pairs:
-            del map_of_PDBs_with_all_residue_pairs[PDB_ID] # We remove PDBs that no longer have all the residue pairs
-        else:
-            map_of_PDBs_with_all_residue_pairs[PDB_ID].extend(new_residue_pair_data[PDB_ID]) # We update the set of residue pairs of the PDBs that have all the residue pairs
+            del map_of_PDBs_with_all_residue_pairs[PDB_ID]
+            del visited_nodes_map[PDB_ID]
+            continue
+        
+        # Connectivity check. If there isn't any residue pair in the new residue pairs that is connected to already visited nodes, then we can also drop this PDB
+        if not any(residue_pair_is_connected_to_visited_nodes(residue_pair, visited_nodes_map[PDB_ID]) for residue_pair in new_residue_pair_data[PDB_ID]):
+            del map_of_PDBs_with_all_residue_pairs[PDB_ID]
+            del visited_nodes_map[PDB_ID]
+            continue
+
+        list_of_residue_pairs = new_residue_pair_data[PDB_ID]
+        map_of_PDBs_with_all_residue_pairs[PDB_ID].extend(list_of_residue_pairs) # Update the list of residue pairs of the PDBs that have all the residue pairs
+        visited_nodes_map[PDB_ID].update(set(itertools.chain.from_iterable(list_of_residue_pairs)))
 
     return
 
@@ -307,10 +331,11 @@ def find_PDBs_with_all_residue_pairs(
     ) -> Dict[str, List[Tuple[str,str]]]:
     """
     """
-    # The algorithm determines which PDBs contain all the residue pairs of a given motif MST by simply iterating over each residue pair 
-    # and calculating the intersection between the current map of PDBs with all residue pairs and the PDBs that contain
-    # the given residue pair that is being checked (i.e iterative intersection).
-    all_pairs_of_residues_to_check = get_all_pairs_of_residues_in_motif_MST(motif_MST)
+    # The algorithm determines which PDBs contain all the residue pairs of a given motif MST by iterating over each residue pair and calculating the 
+    # intersection between the current map of PDBs with all residue pairs and the PDBs that contain the given residue pair that is being 
+    # checked (i.e iterative intersection). A connectivity check between the already determined residue pairs and the new ones is also performed to remove PDBs 
+    # that have the reside pairs but are not connected and therefore don't form a motif (see update_map_of_PDBs_with_all_residue_pairs).
+    all_pairs_of_residues_to_check = get_all_pairs_of_residues_in_motif_MST(motif_MST, order='dfs')
     
     submitted_futures: List[Future[Dict[str, List[Tuple[str,str]]]]] = [
         concurrent_executor.submit(get_PDBs_that_contain_the_residue_pair, pair_of_full_residue_IDs, residue_pair_data, distance_delta_thr, angle_delta_thr, index_folder_path, compression)
@@ -319,7 +344,8 @@ def find_PDBs_with_all_residue_pairs(
 
     map_of_PDBs_with_all_residue_pairs: Dict[str, List[Tuple[str,str]]]
     map_of_PDBs_with_all_residue_pairs_initialized_flag: bool = False
-    for future in as_completed(submitted_futures):
+    visited_nodes_map: Dict[str, Set[str]] # Used for connectivity check
+    for future in submitted_futures:
         if future.exception():
             raise future.exception() # type: ignore
 
@@ -331,13 +357,18 @@ def find_PDBs_with_all_residue_pairs(
         if map_of_PDBs_with_all_residue_pairs_initialized_flag == False:
             map_of_PDBs_with_all_residue_pairs = defaultdict(list, PDBs_that_contain_the_residue_pair)
             map_of_PDBs_with_all_residue_pairs_initialized_flag = True
+            visited_nodes_map = {
+                PDB_ID:set(itertools.chain.from_iterable(list_of_residue_pairs))
+                for PDB_ID, list_of_residue_pairs in map_of_PDBs_with_all_residue_pairs.items()
+            }
 
         else:
             update_map_of_PDBs_with_all_residue_pairs(
                 map_of_PDBs_with_all_residue_pairs, 
-                PDBs_that_contain_the_residue_pair
+                PDBs_that_contain_the_residue_pair,
+                visited_nodes_map
             )
-        
+    
     return map_of_PDBs_with_all_residue_pairs
 
 def add_resname_as_node_attribute(graph: nx.Graph) -> None:
@@ -415,17 +446,23 @@ def solve_motif_MST(
     # Non-parallel version of find_PDBs_with_all_residue_pairs
     map_of_PDBs_with_all_residue_pairs: Dict[str, List[Tuple[str,str]]]
     map_of_PDBs_with_all_residue_pairs_initialized_flag: bool = False
-    for pair_of_full_residue_IDs, residue_pair_data in get_all_pairs_of_residues_in_motif_MST(motif_MST).items():
+    visited_nodes_map: Dict[str, Set[str]] # Used for connectivity check
+    for pair_of_full_residue_IDs, residue_pair_data in get_all_pairs_of_residues_in_motif_MST(motif_MST, order='dfs').items():
         PDBs_that_contain_the_residue_pair = get_PDBs_that_contain_the_residue_pair(pair_of_full_residue_IDs, residue_pair_data, distance_delta_thr, angle_delta_thr, index_folder_path, compression)
         
         if map_of_PDBs_with_all_residue_pairs_initialized_flag == False:
             map_of_PDBs_with_all_residue_pairs = defaultdict(list, PDBs_that_contain_the_residue_pair)
             map_of_PDBs_with_all_residue_pairs_initialized_flag = True
+            visited_nodes_map = {
+                PDB_ID:set(itertools.chain.from_iterable(list_of_residue_pairs))
+                for PDB_ID, list_of_residue_pairs in map_of_PDBs_with_all_residue_pairs.items()
+            }
 
         else:
             update_map_of_PDBs_with_all_residue_pairs(
                 map_of_PDBs_with_all_residue_pairs, 
-                PDBs_that_contain_the_residue_pair
+                PDBs_that_contain_the_residue_pair,
+                visited_nodes_map
             )
 
     # Non-parallel version of filter_out_PDBs_with_unconnected_residue_pairs. Using while loop + popitem()
@@ -457,7 +494,7 @@ def get_PDBs_with_similar_motifs(
     n_motifs_to_solve = sum(1 for _ in get_all_motif_MSTs_generator(reference_motif_MST, max_n_mutated_residues, residue_type_policy, motif_residues_data))
     with ProcessPoolExecutor(max_workers=n_cores) as concurrent_executor:
         # Residue-pair level parallelisation
-        if n_motifs_to_solve < n_cores:
+        if n_motifs_to_solve <= n_cores:
             for motif_MST in get_all_motif_MSTs_generator(reference_motif_MST, max_n_mutated_residues, residue_type_policy, motif_residues_data):
                 PDBs_with_all_residue_pairs = find_PDBs_with_all_residue_pairs(motif_MST, index_folder_path, distance_delta_thr, angle_delta_thr, compression, concurrent_executor)
                 filtered_PDBs_with_all_residue_pairs = filter_out_PDBs_with_unconnected_residue_pairs(PDBs_with_all_residue_pairs, motif_MST, concurrent_executor)
@@ -519,3 +556,61 @@ def search_index_for_PDBs_with_similar_motifs(
     )
 
     return motif_MST, PDBs_with_similar_motifs
+
+
+
+# def update_map_of_PDBs_with_all_residue_pairs(
+#         map_of_PDBs_with_all_residue_pairs: Dict[str, List[Tuple[str,str]]], new_residue_pair_data: Dict[str, List[Tuple[str,str]]]
+#     ) -> None:
+#     """
+#     ...
+#     """
+#     updated_set_of_PDBs_with_all_residue_pairs = map_of_PDBs_with_all_residue_pairs.keys() & new_residue_pair_data.keys() # & = set intersection
+    
+#     for PDB_ID in list(map_of_PDBs_with_all_residue_pairs.keys()):
+#         if PDB_ID not in updated_set_of_PDBs_with_all_residue_pairs:
+#             del map_of_PDBs_with_all_residue_pairs[PDB_ID] # We remove PDBs that no longer have all the residue pairs
+#             continue
+
+#         map_of_PDBs_with_all_residue_pairs[PDB_ID].extend(new_residue_pair_data[PDB_ID]) # We update the set of residue pairs of the PDBs that have all the residue pairs
+
+#     return
+
+# def find_PDBs_with_all_residue_pairs(
+#         motif_MST: nx.Graph, index_folder_path: Path, distance_delta_thr: float, angle_delta_thr: float, compression: str, 
+#         concurrent_executor: ProcessPoolExecutor
+#     ) -> Dict[str, List[Tuple[str,str]]]:
+#     """
+#     """
+#     # The algorithm determines which PDBs contain all the residue pairs of a given motif MST by simply iterating over each residue pair 
+#     # and calculating the intersection between the current map of PDBs with all residue pairs and the PDBs that contain
+#     # the given residue pair that is being checked (i.e iterative intersection).
+#     all_pairs_of_residues_to_check = get_all_pairs_of_residues_in_motif_MST(motif_MST)
+    
+#     submitted_futures: List[Future[Dict[str, List[Tuple[str,str]]]]] = [
+#         concurrent_executor.submit(get_PDBs_that_contain_the_residue_pair, pair_of_full_residue_IDs, residue_pair_data, distance_delta_thr, angle_delta_thr, index_folder_path, compression)
+#         for pair_of_full_residue_IDs, residue_pair_data in all_pairs_of_residues_to_check.items()
+#     ]
+
+#     map_of_PDBs_with_all_residue_pairs: Dict[str, List[Tuple[str,str]]]
+#     map_of_PDBs_with_all_residue_pairs_initialized_flag: bool = False
+#     for future in as_completed(submitted_futures):
+#         if future.exception():
+#             raise future.exception() # type: ignore
+
+#         PDBs_that_contain_the_residue_pair = future.result()
+
+#         # Iterative intersection code. map_of_PDBs_with_all_residue_pairs must be initialized with the PDBs of a 
+#         # residue pair in order to be able to do the iterative intersection, otherwise we would always get an empty set 
+#         # given it is initially empty.
+#         if map_of_PDBs_with_all_residue_pairs_initialized_flag == False:
+#             map_of_PDBs_with_all_residue_pairs = defaultdict(list, PDBs_that_contain_the_residue_pair)
+#             map_of_PDBs_with_all_residue_pairs_initialized_flag = True
+
+#         else:
+#             update_map_of_PDBs_with_all_residue_pairs(
+#                 map_of_PDBs_with_all_residue_pairs, 
+#                 PDBs_that_contain_the_residue_pair
+#             )
+        
+#     return map_of_PDBs_with_all_residue_pairs
